/*
 * Copyright (c) 2008, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <chip.h>
#include <io.h>
#include <serial.h>
#include <sm.h>
#include <util.h>

struct serial_port serial[] = {
	{
		.regs		= (void *)0xffe00c00,
	},
	{
		.regs		= (void *)0xffe01000,
	},
	{
		.regs		= (void *)0xffe01400,
	},
	{
		.regs		= (void *)0xffe01800,
	},
};
unsigned int nr_serial_ports = ARRAY_SIZE(serial);

void enable_hsb_clock(unsigned int id)
{
	u32 mask;

	mask = sm_readl(PM_HSB_MASK);
	sm_writel(PM_HSB_MASK, mask | (1 << id));
}

void enable_pba_clock(unsigned int id)
{
	u32 mask;

	mask = sm_readl(PM_PBA_MASK);
	sm_writel(PM_PBA_MASK, mask | (1 << id));
}

void enable_pbb_clock(unsigned int id)
{
	u32 mask;

	mask = sm_readl(PM_PBB_MASK);
	sm_writel(PM_PBB_MASK, mask | (1 << id));
}

static unsigned long get_pll0_clock_rate(void)
{
	unsigned long rate;
	unsigned long div;
	unsigned long mul;
	u32 control;

	control = sm_readl(PM_PLL0);

	if (!(control & SM_BIT(PLLEN)))
		return 0;

	div = SM_BFEXT(PLLDIV, control) + 1;
	mul = SM_BFEXT(PLLMUL, control) + 1;

	rate = CONFIG_OSC0_HZ;
	rate = (rate + div / 2) / div;
	rate *= mul;

	return rate;
}

static unsigned long get_sync_clock_rate(unsigned int shift)
{
	if (sm_readl(PM_MCCTRL) & SM_BIT(PLLSEL))
		return get_pll0_clock_rate() >> shift;
	else
		return CONFIG_OSC0_HZ >> shift;
}

unsigned long get_cpu_clock_rate(void)
{
	unsigned int shift = 0;
	u32 cksel;

	cksel = sm_readl(PM_CKSEL);
	if (cksel & SM_BIT(CPUDIV))
		shift = SM_BFEXT(CPUSEL, cksel) + 1;

	return get_sync_clock_rate(shift);
}

unsigned long get_hsb_clock_rate(void)
{
	unsigned int shift = 0;
	u32 cksel;

	cksel = sm_readl(PM_CKSEL);
	if (cksel & SM_BIT(HSBDIV))
		shift = SM_BFEXT(HSBSEL, cksel) + 1;

	return get_sync_clock_rate(shift);
}

unsigned long get_pba_clock_rate(void)
{
	unsigned int shift = 0;
	u32 cksel;

	cksel = sm_readl(PM_CKSEL);
	if (cksel & SM_BIT(PBADIV))
		shift = SM_BFEXT(PBASEL, cksel) + 1;

	return get_sync_clock_rate(shift);
}

unsigned long get_pbb_clock_rate(void)
{
	unsigned int shift = 0;
	u32 cksel;

	cksel = sm_readl(PM_CKSEL);
	if (cksel & SM_BIT(PBBDIV))
		shift = SM_BFEXT(PBBSEL, cksel) + 1;

	return get_sync_clock_rate(shift);
}

int paddr_to_cs(unsigned long paddr)
{
	switch ((paddr >> 26) & 0x0f) {
	case 0x00:
		return 0;
	case 0x01:
		return 4;
	case 0x02:
		return 2;
	case 0x03:
		return 3;
	case 0x04:
	case 0x05:
	case 0x06:
	case 0x07:
		return 1;
	case 0x08:
		return 5;
	default:
		return -11;
	}
}
