/*
 * Copyright (c) 2008, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stddef.h>

#include <init.h>
#include <io.h>
#include <flash.h>
#include <smc.h>
#include <util.h>

struct cfi_flash;

struct erase_region {
	u32	sector_size;
	u32	nr_sectors;
};

struct cfi_cmdset {
	unsigned int id;
	const char *name;
	int (*erase_sector)(struct cfi_flash *cfi, unsigned long offset);
	int (*program_sector)(struct cfi_flash *cfi, unsigned long offset,
			      void *data, size_t len);
};

struct cfi_flash {
	struct flash fl;
	const struct cfi_cmdset *cmdset;
	unsigned int nr_erase_regions;
	struct erase_region *erase_region;
};
#define to_cfi_flash(fl) ((struct cfi_flash *)(fl))

struct cfi_qry {
	u8	qry[3];
	u16	p_id;
	u16	p_adr;
	u16	a_id;
	u16	a_adr;
	u8	vcc_min;
	u8	vcc_max;
	u8	vpp_min;
	u8	vpp_max;
	u8	word_write_timeout_typ;
	u8	buf_write_timeout_typ;
	u8	block_erase_timeout_typ;
	u8	chip_erase_timeout_typ;
	u8	word_write_timeout_max;
	u8	buf_write_timeout_max;
	u8	block_erase_timeout_max;
	u8	chip_erase_timeout_max;
	u8	dev_size;
	u16	iface_desc;
	u16	max_buf_write_size;
	u8	num_erase_regions;
	/* erase region info is parsed separately */
} __attribute__((packed));

/* We only support 16-bit data width for now */
#define ADDR_SHIFT	1

#define ADDR_UNLOCK1	0x555
#define ADDR_UNLOCK2	0xaaa

/* Chip identification stuff */
#define MFR_ATMEL	0x001f

#define AT49BV6416	0x00d6
#define AT49BV642D	0x01d6

struct flash_chip {
	const char *name;
	u16 id;
	void (*fixup)(struct cfi_flash *cfi);
};

struct flash_manufacturer {
	const char *name;
	u16 id;
	u16 nr_chips;
	const struct flash_chip *chip;
	void (*fixup)(struct cfi_flash *cfi);
};

static inline void cfi_write16(struct cfi_flash *cfi, unsigned long offset,
			       u16 value)
{
	writew(value, cfi->fl.addr + (offset << ADDR_SHIFT));
}

static inline u16 cfi_read16(struct cfi_flash *cfi, unsigned long offset)
{
	return readw(cfi->fl.addr + (offset << ADDR_SHIFT));
}

int erase_flash(struct flash *flash, unsigned long offset,
		unsigned long len)
{
	struct cfi_flash *cfi = to_cfi_flash(flash);
	unsigned long sector_start;
	unsigned int sector_size;
	unsigned int nr_sectors;
	unsigned int i, j;
	int ret;

	if (offset > flash->size || (offset + len) > flash->size) {
		print("Attempted to erase beyond end of flash", '\n');
		return -1;
	}

	sector_start = 0;
	for (i = 0; i < cfi->nr_erase_regions; i++) {
		sector_size = cfi->erase_region[i].sector_size;
		nr_sectors = cfi->erase_region[i].nr_sectors;

		for (j = 0; j < nr_sectors; j++) {
			if ((sector_start + sector_size) <= offset)
				continue;
			if (sector_start >= (offset + len))
				goto done;

			print("\rErasing...", ' ');
			print_hex(sector_start, 8, '-');
			print_hex(sector_start + sector_size - 1, 8, 0);
			ret = cfi->cmdset->erase_sector(cfi, sector_start);
			if (ret)
				goto fail;

			sector_start += sector_size;
		}
	}

	/* Might get here if we reach the end of the flash */
done:
	print("\rErasing... done                 ", '\n');
	return 0;

fail:
	print("\rErasing... FAILED at offset", ' ');
	print_hex(sector_start, 8, ',');
	print(" status", '=');
	print_hex(ret, 4, '\n');
	return -1;
}

int program_flash(struct flash *flash, unsigned long offset,
		  void *data, unsigned long len)
{
	struct cfi_flash *cfi = to_cfi_flash(flash);
	unsigned long chunk_len;
	unsigned long sector_start;
	unsigned int sector_size;
	unsigned int nr_sectors;
	unsigned int i, j;
	int ret;

	if (offset > flash->size || (offset + len) > flash->size) {
		print("Attempted to program beyond end of flash", '\n');
		return -1;
	}

	sector_start = 0;
	for (i = 0; i < cfi->nr_erase_regions; i++) {
		sector_size = cfi->erase_region[i].sector_size;
		nr_sectors = cfi->erase_region[i].nr_sectors;

		for (j = 0; j < nr_sectors; j++) {
			if ((sector_start + sector_size) <= offset)
				continue;
			if (len == 0)
				goto done;

			chunk_len = sector_size;
			if (sector_start < offset)
				chunk_len -= offset - sector_start;
			if (len < chunk_len)
				chunk_len = len;

			print("\rProgramming...", ' ');
			print_hex(offset, 8, '-');
			print_hex(offset + chunk_len - 1, 8, 0);
			ret = cfi->cmdset->program_sector(cfi, offset,
							  data, chunk_len);
			if (ret)
				goto fail;

			sector_start += sector_size;
			offset += chunk_len;
			data += chunk_len;
			len -= chunk_len;
		}
	}

	/* Might get here if we reach the end of the flash */
done:
	print("\rProgramming... done                 ", '\n');
	return 0;

fail:
	print("\rProgramming... FAILED at offset", ' ');
	print_hex(offset, 8, ',');
	print(" status", '=');
	print_hex(ret, 4, '\n');
	return -1;
}

int verify_flash(struct flash *flash, unsigned long offset,
		 void *data, unsigned long len)
{
	u16 *src = data;
	u16 *dst = flash->addr + offset;
	u16 expected, actual;

	print("Verifying...", ' ');

	len = (len + 1) & ~1UL;
	for(; len; len -= 2, src++, dst++, offset += 2) {
		expected = *src;
		actual = *dst;

		if (expected != actual) {
			print("FAILED at offset", ' ');
			print_hex(offset, 8, '\n');
			print("Expected:", ' ');
			print_hex(expected, 4, ',');
			print(" actual:", ' ');
			print_hex(actual, 4, '\n');
			return -1;
		}
	}

	print("done", '\n');

	return 0;
}

static int cfi_amd_erase_sector(struct cfi_flash *cfi, unsigned long offset)
{
	u16 this, prev;
	unsigned int timeout;

	cfi_write16(cfi, ADDR_UNLOCK1, 0xaa);
	cfi_write16(cfi, ADDR_UNLOCK2, 0x55);
	cfi_write16(cfi, ADDR_UNLOCK1, 0x80);
	cfi_write16(cfi, ADDR_UNLOCK1, 0xaa);
	cfi_write16(cfi, ADDR_UNLOCK2, 0x55);
	writew(0x30, cfi->fl.addr + offset);

	prev = readw(cfi->fl.addr + offset);

	for (timeout = CONFIG_SECTOR_ERASE_TIMEOUT; timeout; timeout--) {
		udelay(1000);
		this = readw(cfi->fl.addr + offset);

		if (!((this ^ prev) & 0x40))
			break;
		if (this & 0x20) {
			prev = readw(cfi->fl.addr + offset);
			this = readw(cfi->fl.addr + offset);
			break;
		}
		prev = this;
	}

	if (timeout != 0 && !((this ^ prev) & 0x40) && this == 0xffff)
		return 0;

	/* Failed. Put the device back in read mode */
	cfi_write16(cfi, 0, 0xf0);

	return this;
}

static int cfi_amd_program_sector(struct cfi_flash *cfi, unsigned long offset,
				  void *data, size_t len)
{
	u16 *src = data;
	u16 *dst = cfi->fl.addr + offset;
	unsigned int timeout;
	u16 this, prev;

	if ((unsigned long)data & 1 || len & 1)
		panic("cfi: buffer and/or len not halfword-aligned\n");

	for (; len; len -= 2, src++, dst++, offset += 2) {
		u16 word = *src;

		cfi_write16(cfi, ADDR_UNLOCK1, 0xaa);
		cfi_write16(cfi, ADDR_UNLOCK2, 0x55);
		cfi_write16(cfi, ADDR_UNLOCK1, 0xa0);
		writew(word, dst);

		prev = readw(dst);

		for (timeout = (CONFIG_WORD_PROGRAM_TIMEOUT + 5) / 10;
		     timeout; timeout--) {
			udelay(10);
			this = readw(dst);

			if (!((this ^ prev) & 0x40))
				break;
			if (this & 0x20) {
				prev = readw(dst);
				this = readw(dst);
				break;
			}
			prev = this;
		}
		if (timeout == 0 || ((this ^ prev) & 0x40) || this != word) {
			return this;
		}
	}

	return 0;
}

static const struct cfi_cmdset cmdset_table[] = {
	{
		.id		= 2,
		.name		= "AMD/Fujitsu",
		.erase_sector	= cfi_amd_erase_sector,
		.program_sector	= cfi_amd_program_sector,
	},
};

static void read_cfi_block(struct cfi_flash *cfi, unsigned long offset,
			   void *buf, size_t len)
{
	u8 *p = buf;

	cfi_write16(cfi, 0x55, 0x98);
	while (len--)
		*p++ = cfi_read16(cfi, offset++);

	cfi_write16(cfi, 0, 0xf0);
	cfi_write16(cfi, 0, 0xff);
}

/* Fixups go here */

static void fixup_at49bv6416(struct cfi_flash *cfi)
{
	struct erase_region tmp;
	unsigned long offset;
	unsigned int i, j;

	/*
	 * Erase regions are supposed to start at the lowest address
	 * for bottom-boot devices. The AT49BV6416 is different. Swap
	 * the erase regions to avoid confusion later.
	 */
	memcpy(&tmp, &cfi->erase_region[0], sizeof(struct erase_region));
	memcpy(&cfi->erase_region[0], &cfi->erase_region[1],
	       sizeof(struct erase_region));
	memcpy(&cfi->erase_region[1], &tmp, sizeof(struct erase_region));

	/*
	 * The chip also features a "soft unlock" operation that needs
	 * to be performed before programming. Do it now.
	 */
	offset = 0;
	for (i = 0; i < cfi->nr_erase_regions; i++) {
		for (j = 0; j < cfi->erase_region[i].nr_sectors; j++) {
			cfi_write16(cfi, ADDR_UNLOCK1, 0xaa);
			writew(0x70, cfi->fl.addr + offset);
			offset += cfi->erase_region[i].sector_size;
		}
	}
}

static const struct flash_chip atmel_chips[] = {
	{
		.name		= "AT49BV6416",
		.id		= AT49BV6416,
		.fixup		= fixup_at49bv6416,
	},
	{
		.name		= "AT49BV642D",
		.id		= AT49BV642D,
	},
};

static const struct flash_manufacturer manufacturer[] = {
	{
		.name		= "Atmel",
		.id		= MFR_ATMEL,
		.nr_chips	= ARRAY_SIZE(atmel_chips),
		.chip		= atmel_chips,
	},
};

static int identify_chip(struct cfi_flash *cfi)
{
	const struct flash_manufacturer *mfr = NULL;
	const struct flash_chip *chip = NULL;
	unsigned long total_erase_region_size;
	u32 region_info;
	unsigned int cmdset_id;
	unsigned int i;
	struct cfi_qry qry;

	for (i = 0; i < ARRAY_SIZE(manufacturer); i++) {
		if (manufacturer[i].id == cfi->fl.mfr_id) {
			mfr = &manufacturer[i];
			break;
		}
	}

	if (mfr) {
		for (i = 0; i < mfr->nr_chips; i++) {
			if (mfr->chip[i].id == cfi->fl.dev_id) {
				chip = &mfr->chip[i];
				break;
			}
		}
	}

	if (!chip) {
		print("cfi: Unknown chip", ' ');
		print_hex(cfi->fl.mfr_id, 4, ':');
		print_hex(cfi->fl.dev_id, 4, '\n');
		return -1;
	}

	print(mfr->name, ' ');
	print(chip->name, ' ');
	print("found at address ", 0);
	print_hex(cfi->fl.phys_addr, 8, '\n');

	read_cfi_block(cfi, 0x10, &qry, sizeof(qry));

	if (qry.qry[0] != 'Q' || qry.qry[1] != 'R' || qry.qry[2] != 'Y') {
		print("cfi: invalid CFI QRY block:", ' ');
		print_hex(qry.qry[0], 2, ' ');
		print_hex(qry.qry[1], 2, ' ');
		print_hex(qry.qry[2], 2, '\n');
		return -1;
	}

	cfi->fl.size = 1 << qry.dev_size;
	cfi->nr_erase_regions = qry.num_erase_regions;
	cfi->erase_region = zalloc(sizeof(struct erase_region)
				   * cfi->nr_erase_regions);
	for (i = 0; i < cfi->nr_erase_regions; i++) {
		u32 sector_size, nr_sectors;

		read_cfi_block(cfi, 0x10 + sizeof(qry) + i * 4,
			       &region_info, sizeof(region_info));
		region_info = le32_to_cpu(region_info);
		if (region_info & 0xffff0000)
			sector_size = 256 * (region_info >> 16);
		else
			sector_size = 128;
		nr_sectors = (region_info & 0xffff) + 1;

		cfi->erase_region[i].sector_size = sector_size;
		cfi->erase_region[i].nr_sectors = nr_sectors;
	}

	cmdset_id = le16_to_cpu(qry.p_id);
	for (i = 0; i < ARRAY_SIZE(cmdset_table); i++) {
		if (cmdset_table[i].id == cmdset_id) {
			cfi->cmdset = &cmdset_table[i];
			break;
		}
	}

	if (!cfi->cmdset) {
		print("cfi: unknown command set", ' ');
		print_hex(cmdset_id, 4, '\n');
		return -1;
	}

	/* Run manufacturer- and chip-specific fixups, if any */
	if (mfr->fixup)
		mfr->fixup(cfi);
	if (chip->fixup)
		chip->fixup(cfi);

	print("cfi: using", ' ');
	print(cfi->cmdset->name, ' ');
	print("command set", '\n');

	/*
	 * Sanity-check the erase regions. The should add up to the
	 * total size of the device.
	 */
	print("cfi:", ' ');
	print_dec(cfi->nr_erase_regions, ' ');
	print("erase regions (total size:", ' ');
	print_dec(cfi->fl.size, ' ');
	print("bytes)", '\n');

	total_erase_region_size = 0;
	for (i = 0; i < cfi->nr_erase_regions; i++) {
		u32 sector_size, nr_sectors;

		sector_size = cfi->erase_region[i].sector_size;
		nr_sectors = cfi->erase_region[i].nr_sectors;
		total_erase_region_size += nr_sectors * sector_size;

		print("  ", 0);
		print_dec(i, ' ');
		print_dec(nr_sectors, ' ');
		print("sectors,", ' ');
		print_dec(sector_size, ' ');
		print("bytes each", '\n');
	}

	if (total_erase_region_size != cfi->fl.size) {
		print("cfi: erase region sanity check failed", '\n');
		return -1;
	}

	return 0;
}

static const struct smc_config flash_config = {
	.ncs_read_setup         = 0,
	.nrd_setup              = 40,
	.ncs_write_setup        = 0,
	.nwe_setup              = 10,

	.ncs_read_pulse         = 80,
	.nrd_pulse              = 40,
	.ncs_write_pulse        = 65,
	.nwe_pulse              = 55,

	.read_cycle             = 120,
	.write_cycle            = 120,

	.bus_width              = 2,
	.nrd_controlled         = 1,
	.nwe_controlled         = 1,
	.byte_write             = 1,
};

struct flash *init_flash(unsigned long phys_addr)
{
	struct cfi_flash *cfi;

	smc_set_configuration(phys_addr, &flash_config);

	cfi = zalloc(sizeof(struct cfi_flash));

	cfi->fl.phys_addr = phys_addr;
	cfi->fl.addr = phys_to_uncached(phys_addr);

	/* Make sure that we are in Read Mode */
	cfi_write16(cfi, 0, 0xf0);

	/* Enter Product ID Mode */
	cfi_write16(cfi, ADDR_UNLOCK1, 0xaa);
	cfi_write16(cfi, ADDR_UNLOCK2, 0x55);
	cfi_write16(cfi, ADDR_UNLOCK1, 0x90);
	cfi->fl.mfr_id = cfi_read16(cfi, 0);
	cfi->fl.dev_id = cfi_read16(cfi, 1);

	/* Go back to Read Mode (AMD and Intel use different commands) */
	cfi_write16(cfi, 0, 0xf0);
	cfi_write16(cfi, 0, 0xff);

	if (identify_chip(cfi))
		panic("CFI flash identification failed\n");

	return &cfi->fl;
}
