/*
 * Copyright (c) 2008, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <chip.h>
#include <flash.h>
#include <init.h>
#include <serial.h>
#include <sysreg.h>
#include <util.h>

static unsigned long cpu_hz;
static unsigned char *brk = _end;

void *malloc(unsigned int size)
{
	unsigned long sp;
	void *ret;

	asm("mov %0, sp" : "=r"(sp));
	if ((unsigned long)brk + size + 128 >= sp)
		panic("Memory allocation failed");

	ret = brk;
	brk += (size + 3) & ~3UL;
	return ret;
}

void *zalloc(unsigned int size)
{
	void *p;

	p = malloc(size);
	memset(p, 0, size);
	return p;
}

void print(const char *s, int ch)
{
	serial_puts(s);
	if (ch)
		serial_putchar(ch);
}

void print_dec(unsigned long n, int ch)
{
	char *p;
	char buf[16];

	p = &buf[15];
	*p = '\0';

	if (!n)
		*(--p) = '0';

	while (n) {
		*(--p) = '0' + n % 10;
		n /= 10;
	}

	serial_puts(p);
	if (ch)
		serial_putchar(ch);
}

void print_hex(unsigned long n, unsigned int width, int ch)
{
	char *p;
	int w = width;
	int c;
	char buf[16];

	if (width > 8)
		width = 8;

	p = &buf[15];
	*p = '\0';

	while (n) {
		c = n & 0xf;
		n >>= 4;
		if (c < 10)
			c += '0';
		else
			c += 'a' - 10;

		*(--p) = c;
		w--;
	}

	while (w-- > 0)
		*(--p) = '0';

	*(--p) = 'x';
	*(--p) = '0';

	serial_puts(p);
	if (ch)
		serial_putchar(ch);
}

void __noreturn panic(const char *msg)
{
	print("Panic:", ' ');
	print(msg, 0);
	for (;;);
}

void udelay(unsigned long us)
{
	unsigned long now;
	unsigned long end;

	now = sysreg_read(COUNT);
	end = now + (us * ((cpu_hz + 99999) / 100000)) / 10;
	
	while (now > end)
		now = sysreg_read(COUNT);
	while (end > now)
		now = sysreg_read(COUNT);
}

int main(void)
{
	struct flash *flash;
	unsigned long image_size;
	int ret;
	char c;

	init_serial();
	print("AVR32 Flash upgrade utility version "
	      VERSION_STRING "\n\n", 0);
	cpu_hz = get_cpu_clock_rate();
	flash = init_flash(CONFIG_FLASH_BASE);

	image_size = image_end - image_start;

	print("Going to copy", ' ');
	print_dec(image_size, ' ');
	print("bytes to offset", ' ');
	print_hex(CONFIG_FLASH_OFFSET, 8, ' ');
	print("in flash", '\n');

	/* Empty the USART RX buffer */
	while (serial_poll())
		serial_getchar();

	print("Press `y' to continue, or any other key to abort", '\n');
	c = serial_getchar();

	if (c != 'y' && c != 'Y') {
		print("Aborted", '\n');
		return 1;
	}

retry:
	ret = 0;
	if (erase_flash(flash, CONFIG_FLASH_OFFSET, image_size)
	    || program_flash(flash, CONFIG_FLASH_OFFSET,
			     image_start, image_size)
	    || verify_flash(flash, CONFIG_FLASH_OFFSET,
			    image_start, image_size)) {
		ret = 1;
		print("Flash programming failed. Try again?", '\n');
		print("Press `y' to try again, or `n' to abort", '\n');
		do {
			c = serial_getchar();
		} while (c != 'y' && c != 'Y'
			 && c != 'n' && c != 'N');

		if (c == 'y' || c == 'Y')
			goto retry;
	}

	if (ret == 0)
		print("Flash upgrade successful. "
		      "Please press reset or cycle power.", '\n');
	else
		print("Flash upgrade NOT successful.", '\n');

	return ret;
}
